# Copyright (c) Facebook, Inc. and its affiliates.
import enum

import numpy as np

import nle
from nle.env import base
from nle import nethack


TASK_ACTIONS = tuple(
    [nethack.MiscAction.MORE]
    + list(nethack.CompassDirection)
    + list(nethack.CompassDirectionLonger)
    + list(nethack.MiscDirection)
    + [nethack.Command.KICK, nethack.Command.EAT, nethack.Command.SEARCH]
)


class NetHackScore(base.NLE):
    """Environment for "score" task.

    The task is an augmentation of the standard NLE task. The return function is
    defined as:
    :math:`\text{score}_t - \text{score}_{t-1} + \text{TP}`,
    where the :math:`\text{TP}` is a time penalty that grows with the amount of
    environment steps that do not change the state (such as navigating menus).

    Args:
        penalty_mode (str): name of the mode for calculating the time step
            penalty. Can be ``constant``, ``exp``, ``square``, ``linear``.
            Defaults to ``constant``.
        penalty_step (float): constant applied to amount of frozen steps.
            Defaults to -0.01.
        penalty_time (float): constant applied to amount of frozen steps.
            Defaults to -0.0.

    """

    def __init__(
        self,
        *args,
        penalty_mode="constant",
        penalty_step: float = -0.01,
        penalty_time: float = -0.0,
        **kwargs,
    ) -> None:
        self.penalty_mode = penalty_mode
        self.penalty_step = penalty_step
        self.penalty_time = penalty_time

        self._frozen_steps = 0

        actions = kwargs.pop("actions", TASK_ACTIONS)
        super().__init__(*args, actions=actions, **kwargs)

    def _get_time_penalty(self, last_response, response):
        blstats_old = last_response.Blstats()
        blstats_new = response.Blstats()

        if blstats_old is None or blstats_new is None:
            return 0
        else:
            old_time = blstats_old.Time()
            new_time = blstats_new.Time()

            if old_time == new_time:
                self._frozen_steps += 1
            else:
                self._frozen_steps = 0

            penalty = 0
            if self.penalty_mode == "constant":
                if self._frozen_steps > 0:
                    penalty += self.penalty_step
            elif self.penalty_mode == "exp":
                penalty += 2 ** self._frozen_steps * self.penalty_step
            elif self.penalty_mode == "square":
                penalty += self._frozen_steps ** 2 * self.penalty_step
            elif self.penalty_mode == "linear":
                penalty += self._frozen_steps * self.penalty_step
            else:  # default
                penalty += self.penalty_step
            penalty += (new_time - old_time) * self.penalty_time
            return penalty

    def _reward_fn(self, last_response, response, end_status):
        """Score delta, but with added a state loop penalty."""
        score_diff = super()._reward_fn(last_response, response, end_status)
        time_penalty = self._get_time_penalty(last_response, response)
        return score_diff + time_penalty


class NetHackStaircase(NetHackScore):
    """Environment for "staircase" task.

    This task requires the agent to get on top of a staircase down (>).
    The reward function is :math:`I + \text{TP}`, where :math:`I` is 1 if the
    task is successful, and 0 otherwise, and :math:`\text{TP}` is the time step
    function as defined by `NetHackScore`.
    """

    class StepStatus(enum.IntEnum):
        RUNNING = 0
        DEATH = 1
        TASK_SUCCESSFUL = 2

    def _is_episode_end(self, response) -> None:
        internal = response.Internal()
        if internal and internal.StairsDown():
            return self.StepStatus.TASK_SUCCESSFUL
        return self.StepStatus.RUNNING

    def _reward_fn(self, last_response, response, end_status):
        time_penalty = self._get_time_penalty(last_response, response)
        if end_status == self.StepStatus.TASK_SUCCESSFUL:
            reward = 1
        else:
            reward = 0
        return reward + time_penalty


class NetHackStaircasePet(NetHackStaircase):
    """Environment for "staircase-pet" task.

    This task requires the agent to get on top of a staircase down (>), while
    having their pet next to it. See `NetHackStaircase` for the reward function.
    """

    def _is_episode_end(self, response) -> None:
        internal = response.Internal()
        if internal and internal.StairsDown():
            obs = response.Observation()
            s = response.Blstats()
            if obs and s:
                glyphs = (
                    obs.Glyphs()
                    .DataAsNumpy()
                    .view(np.int16)
                    .reshape(base.DUNGEON_SHAPE)
                )
                x = s.CursX()
                y = s.CursY()
                neighbors = glyphs[y - 1 : y + 2, x - 1 : x + 2].reshape(-1).tolist()
                # TODO: vectorize
                for glyph in neighbors:
                    if nethack.glyph_is_pet(glyph):
                        return self.StepStatus.TASK_SUCCESSFUL
        return self.StepStatus.RUNNING


class NetHackOracle(NetHackStaircase):
    """Environment for "oracle" task.

    This task requires the agent to reach the oracle (by standing next to it).
    See `NetHackStaircase` for the reward function.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.oracle_glyph = None
        for glyph in range(nethack.GLYPH_MON_OFF, nethack.GLYPH_PET_OFF):
            if nethack.glyph_to_mon(glyph).mname == "Oracle":
                self.oracle_glyph = glyph
                break
        assert self.oracle_glyph is not None

    def _is_episode_end(self, response) -> None:
        internal = response.Internal()
        if internal:
            obs = response.Observation()
            s = response.Blstats()
            if obs and s:
                glyphs = (
                    obs.Glyphs()
                    .DataAsNumpy()
                    .view(np.int16)
                    .reshape(base.DUNGEON_SHAPE)
                )
                x = s.CursX()
                y = s.CursY()
                neighbors = glyphs[y - 1 : y + 2, x - 1 : x + 2].reshape(-1).tolist()
                # TODO: vectorize
                if self.oracle_glyph in neighbors:
                    return self.StepStatus.TASK_SUCCESSFUL
        return self.StepStatus.RUNNING


class NetHackGold(NetHackScore):
    """Environment for the "gold" task.

    The task is similar to the one defined by `NetHackScore`, but the reward
    uses changes in the amount of gold collected by the agent, rather than the
    score.

    The agent will pickup gold automatically by walking on top of it.
    """

    def __init__(
        self, *args, **kwargs,
    ):
        options = kwargs.pop(
            "options",
            (
                "windowtype:rl",
                "color",
                "showexp",
                "nobones",
                "autopickup",
                "pickup_types:$",
            ),
        )

        super().__init__(*args, options=options, **kwargs)

    def _reward_fn(self, last_response, response, end_status):
        """Difference between previous gold and new gold."""
        del end_status  # Unused

        old_gold = base._get(last_response, "Blstats.gold", 0)
        gold = base._get(response, "Blstats.gold", old_gold)
        time_penalty = self._get_time_penalty(last_response, response)

        return gold - old_gold + time_penalty


# FIXME: the way the reward function is currently structured means the
# agents gets a penalty of -1 every other step (since the
# uhunger increases by that)
# thus the step penalty becomes irrelevant
class NetHackEat(NetHackScore):
    """Environment for the "eat" task.

    The task is similar to the one defined by `NetHackScore`, but the reward
    uses positive changes in the character's hunger level (e.g. by consuming
    comestibles or monster corpses), rather than the score.
    """

    def _reward_fn(self, last_response, response, end_status):
        """Difference between previous hunger and new hunger."""
        del end_status  # Unused

        old_hunger = base._get(last_response, "You.uhunger", 0)
        hunger = base._get(response, "You.uhunger", old_hunger)

        reward = max(0, hunger - old_hunger)

        time_penalty = self._get_time_penalty(last_response, response)

        return reward + time_penalty


class NetHackScout(NetHackScore):
    """Environment for the "scout" task.

    The task is similar to the one defined by `NetHackScore`, but the score is
    defined by the changes in glyphs discovered by the agent.
    """

    def reset(self, *args, **kwargs):
        self.dungeon_explored = {}
        return super().reset(*args, **kwargs)

    def _reward_fn(self, last_response, response, end_status):
        del end_status  # Unused

        reward = 0
        internal = response.Internal()
        if internal:
            obs = response.Observation()
            s = response.Blstats()
            if obs and s:
                glyphs = (
                    obs.Glyphs()
                    .DataAsNumpy()
                    .view(np.int16)
                    .reshape(base.DUNGEON_SHAPE)
                )
                dlevel = response.You().Uz(nle.fbs.DLevel.DLevel())
                dungeon_num = dlevel.Dnum()
                dungeon_level = dlevel.Dlevel()

                key = (dungeon_num, dungeon_level)
                if glyphs is not None:
                    explored = (glyphs != 0).sum()
                    explored_old = 0
                    if key in self.dungeon_explored:
                        explored_old = self.dungeon_explored[key]
                    reward = explored - explored_old
                    self.dungeon_explored[key] = explored
        time_penalty = self._get_time_penalty(last_response, response)
        return reward + time_penalty



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import os
import select
import struct
import termios
import time

parser = argparse.ArgumentParser()
parser.add_argument(
    "-1",
    "--no_input",
    action="store_true",
    help="Use ttyrec (not ttyrec2) format without input data",
)
parser.add_argument(
    "-n",
    "--no_wait",
    action="store_true",
    help="Ignore timestamp data and don't wait between frames",
)
parser.add_argument(
    "-s", "--speed", default=1.0, type=float, help="Set playback speed multiplier"
)
parser.add_argument(
    "-p", "--peek", action="store_true", help="Peek mode (like tail -f)"
)
parser.add_argument(
    "filename", default="-", type=str, nargs="?", help="tty record file, or - for stdin"
)
parser.add_argument("--start", default=0, type=int, help="Start at a specific frame")
parser.add_argument(
    "--end", default=float("inf"), type=int, help="Quit after a specific frame count"
)


def wait(diff, speed, drift=0.0):
    jump = 0

    if FLAGS.no_wait:
        return speed, drift, jump

    start = time.time()

    diff /= speed
    diff = max(diff - drift, 0.0)

    rlist, _, _ = select.select((0,), (), (), diff)
    if rlist:
        c = os.read(0, 1)
        if c in (b"+", b"f"):
            speed *= 2
        elif c in (b"-", b"s"):
            speed /= 2
        elif c == b"1":
            speed = 1
        elif c == b"h":
            jump = -1
        elif c == b"l":
            jump = 1
        elif c == b" ":
            select.select((0,), (), ())
        drift = 0.0
    else:
        stop = time.time()
        if diff == 0.0:
            diff = -drift
        drift = (stop - start) - diff
    return speed, drift, jump


def read_header(fd, peek=False, no_input=False):
    while True:
        if no_input:
            header = os.read(fd, 12)
        else:
            header = os.read(fd, 13)

        if not header:
            if peek:
                # Never return, just wait for more data.
                time.sleep(0.25)
                continue
            return

        if no_input:
            sec, usec, length = struct.unpack("<iii", header)
            channel = 0
        else:
            sec, usec, length, channel = struct.unpack("<iiiB", header)

        if sec < 0 or usec < 0 or length < 1 or channel not in (0, 1):
            raise IOError("Illegal header %s" % ((sec, usec, length, channel),))
        timestamp = sec + usec * 1e-6
        yield timestamp, length, channel


CLRCODE = b"\033[2J"


def process(fd):
    speed = FLAGS.speed
    drift = 0.0
    prev = None
    jump = 0

    # Store the header positions before clear screen commands,
    # the timestamp before.
    clrscreen = []

    lastpos = 0
    frame = 0

    for timestamp, length, channel in read_header(
        fd, peek=FLAGS.peek, no_input=FLAGS.no_input
    ):
        data = os.read(fd, length)
        frame += 1

        if frame < FLAGS.start:
            continue

        if channel == 1:  # Input channel.
            continue

        if jump == 0 and prev is not None:
            speed, drift, jump = wait(timestamp - prev, speed, drift)

        offset = data.find(CLRCODE)
        if offset >= 0:
            clrscreen.append((lastpos, prev))
            if jump > 0:
                jump = 0

        lastpos = os.lseek(fd, 0, os.SEEK_CUR)

        prev = timestamp

        if jump > 0:
            continue
        if jump < 0:
            jump = 0

            if not clrscreen:
                clrscreen.append((0, None))

            while prev is not None and timestamp - prev <= 0.005:
                lastpos, prev = clrscreen.pop()
            if lastpos:
                prev = timestamp

            os.lseek(fd, lastpos, os.SEEK_SET)
            continue

        os.write(1, data)

        if frame > FLAGS.end:
            return


def main():
    global FLAGS
    FLAGS = parser.parse_args()

    if FLAGS.filename == "-":
        fd = os.dup(0)
        os.dup2(1, 0)
    else:
        fd = os.open(FLAGS.filename, os.O_RDONLY)

    old = termios.tcgetattr(0)
    new = termios.tcgetattr(0)

    # Set to unbuffered, no echo.
    new[3] &= ~(termios.ICANON | termios.ECHO | termios.ECHONL)  # lflags
    termios.tcsetattr(0, termios.TCSANOW, new)

    try:
        if FLAGS.peek:
            # Skip all previous data.
            for _, length, _ in read_header(fd, peek=False):
                os.lseek(fd, length, os.SEEK_CUR)
            FLAGS.no_wait = True
        process(fd)
    except KeyboardInterrupt:
        pass
    finally:
        termios.tcsetattr(0, termios.TCSANOW, old)
        os.close(fd)


if __name__ == "__main__":
    main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Internal(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsInternal(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Internal()
        x.Init(buf, n + offset)
        return x

    # Internal
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Internal
    def DeepestLevReached(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Internal
    def CallStack(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Internal
    def CallStackLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Internal
    def CallStackIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Internal
    def KillerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Internal
    def Xwaitforspace(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Internal
    def StairsDown(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def InternalStart(builder): builder.StartObject(5)
def InternalAddDeepestLevReached(builder, deepestLevReached): builder.PrependInt8Slot(0, deepestLevReached, 0)
def InternalAddCallStack(builder, callStack): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(callStack), 0)
def InternalStartCallStackVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def InternalAddKillerName(builder, killerName): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(killerName), 0)
def InternalAddXwaitforspace(builder, xwaitforspace): builder.PrependBoolSlot(3, xwaitforspace, 0)
def InternalAddStairsDown(builder, stairsDown): builder.PrependBoolSlot(4, stairsDown, 0)
def InternalEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class NDArray(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsNDArray(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NDArray()
        x.Init(buf, n + offset)
        return x

    # NDArray
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # NDArray
    def Shape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # NDArray
    def ShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # NDArray
    def ShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NDArray
    def ShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # NDArray
    def Dtype(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # NDArray
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # NDArray
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # NDArray
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NDArray
    def DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def NDArrayStart(builder): builder.StartObject(3)
def NDArrayAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def NDArrayStartShapeVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def NDArrayAddDtype(builder, dtype): builder.PrependInt8Slot(1, dtype, 0)
def NDArrayAddData(builder, data): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def NDArrayStartDataVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def NDArrayEnd(builder): return builder.EndObject()



# Copyright (c) Facebook, Inc. and its affiliates.
"""Debugging script for the NLE dev team."""

import random
import timeit

from nle import nethack

ACTIONS = [nethack.MiscAction.MORE]
ACTIONS += list(nethack.CompassDirection)
ACTIONS += list(nethack.CompassDirectionLonger)


def play(game, print_tombstone=True):
    observation = game.reset()

    steps = 0

    while True:
        ch = random.choice(ACTIONS)
        last_observation = observation
        observation, done, unused_info = game.step(ch)
        steps += 1
        if steps > 10000:
            print("Game abandonned after", steps, "steps.")
            return steps
        if done:
            if last_observation.ProgramState().Gameover():
                # Print tombstone.
                if last_observation.WindowsLength() < 1:
                    return steps
                window = last_observation.Windows(1)

                if print_tombstone:
                    for i in range(window.StringsLength()):
                        print(window.Strings(i).decode("ascii"))
                    killer_name = last_observation.Internal().KillerName()
                    if killer_name:
                        print(killer_name.decode("utf-8"))

            return steps


def main():
    episodes = 0
    steps = 0

    game = nethack.NetHack(archivefile="random.zip")

    start = timeit.default_timer()
    while True:
        steps_delta = play(game, False)
        time_delta = timeit.default_timer() - start

        episodes += 1
        steps += steps_delta

        print(
            "Episde: %i. Steps: %i. SPS: %f"
            % (episodes, steps, steps_delta / time_delta)
        )
        start = timeit.default_timer()


if __name__ == "__main__":
    main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class InventoryItem(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsInventoryItem(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = InventoryItem()
        x.Init(buf, n + offset)
        return x

    # InventoryItem
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # InventoryItem
    def Glyph(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # InventoryItem
    def Str(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # InventoryItem
    def Letter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # InventoryItem
    def ObjectClass(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # InventoryItem
    def ObjectClassName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def InventoryItemStart(builder): builder.StartObject(5)
def InventoryItemAddGlyph(builder, glyph): builder.PrependInt16Slot(0, glyph, 0)
def InventoryItemAddStr(builder, str): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(str), 0)
def InventoryItemAddLetter(builder, letter): builder.PrependInt8Slot(2, letter, 0)
def InventoryItemAddObjectClass(builder, objectClass): builder.PrependInt8Slot(3, objectClass, 0)
def InventoryItemAddObjectClassName(builder, objectClassName): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(objectClassName), 0)
def InventoryItemEnd(builder): return builder.EndObject()



# Copyright (c) Facebook, Inc. and its affiliates.
import concurrent.futures
import io
import unittest
import os
import struct
import sys
import time

from nle.nethack import ptyprocess


def _proc_target():
    for i in range(1000000):
        sys.stdout.write(str(i))
        sys.stdout.flush()
        time.sleep(0.1)


def read_ttyrec_header(f):
    while True:
        header = f.read(13)
        if not header:
            return
        sec, usec, length, channel = struct.unpack("<iiiB", header)
        if sec < 0 or usec < 0 or length < 1 or channel not in (0, 1):
            raise IOError("Illegal header %s" % ((sec, usec, length, channel),))
        timestamp = sec + usec * 1e-6
        yield timestamp, length, channel


class PtyProcess(unittest.TestCase):
    def test_simple(self):
        calls = 0

        def recordclosefn(filename):
            nonlocal calls
            calls += 1
            self.assertTrue(os.path.exists(filename))
            os.unlink(filename)

        p = ptyprocess.PtyProcess(_proc_target, recordclosefn=recordclosefn)
        p.fork()

        time.sleep(0.5)
        del p
        self.assertEqual(calls, 1)

    def test_recordfile(self):
        future = concurrent.futures.Future()

        def recordclosefn(filename):
            self.assertTrue(os.path.exists(filename))
            future.set_result(filename)

        p = ptyprocess.PtyProcess(_proc_target, recordclosefn=recordclosefn)
        p.fork()

        for _ in range(3):
            p.write(b"x")

        time.sleep(0.5)
        self.assertFalse(future.done())
        del p
        self.assertTrue(future.done())

        streams = [io.BytesIO(), io.BytesIO()]

        filename = future.result()
        with open(filename, "rb") as f:
            for _, length, channel in read_ttyrec_header(f):
                streams[channel].write(f.read(length))

        self.assertEqual(streams[0].getvalue().count(b"x"), 3)
        for i, c in enumerate(streams[0].getvalue().replace(b"x", b"")):
            self.assertEqual(c, ord(str(i)))
        self.assertEqual(streams[1].getvalue(), b"xxx")

        os.unlink(filename)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# List of environment variables used:
#
#  NLE_PACKAGE_NAME
#    Prefix of the generated package (defaults to "nle").
#
#  NLE_BUILD_RELEASE
#    If set, builds wheel (s)dist such as to prepare it for upload to PyPI.
#
#  NLE_BUILD_FASTER
#    If set, runs `make -j` instead of `make`, when building NetHack.
#    NOTE: we don't make use of MAKEFLAGS (which would be The Right Thing To
#          Do™) because `make install` is not robust against it on all systems,
#          and this is an easier patch towards enabling less annoying
#          development cycles.
import sysconfig
import shutil
import os
import sys
import subprocess
import setuptools
import setuptools.command.build_ext
import setuptools.command.develop
import setuptools.command.install
import distutils.command.build


def make_install_nethack():
    if sys.platform == "darwin":
        hints = "hints/macosx"
        setupcmd = ["sh", "setup.sh", "hints/macosx"]
    else:
        hints = "hints/linux"
    setupcmd = ["sh", "setup.sh", hints]

    if subprocess.call(setupcmd, cwd="sys/unix") != 0:
        print("Failed to generate NetHack makefiles with {}".format(hints))
        sys.exit(-1)

    print("Generated NetHack makefiles with {}.".format(hints))

    build_env = os.environ.copy()

    if "PREFIX" not in build_env:
        build_env["PREFIX"] = sysconfig.get_config_var("base")

    cpath = build_env.get("CPATH", "").split(":")
    cpath.append(os.path.join(build_env["PREFIX"], "include"))
    build_env["CPATH"] = ":".join(cpath)

    makecmd = ["make"]
    if os.getenv("NLE_BUILD_FASTER") is not None:
        makecmd += ["-j"]

    if subprocess.call(makecmd, env=build_env) != 0:
        print("Failed to build NetHack")
        sys.exit(-1)
    print("Successfully built NetHack")

    if subprocess.call(["make", "install"], env=build_env) != 0:
        print("Failed to install NetHack")
        sys.exit(-1)
    print("Successfully installed NetHack")


def build_fbs():
    source_path = "win/rl"
    cmd = ["flatc", "-o", "win/rl", "--python", "win/rl/message.fbs"]
    err = subprocess.call(cmd)
    if err != 0:
        print("Error while building fbs python interface: {}".format(err))
        sys.exit(-1)
    print("Successfully built fbs files from {}".format(source_path))


def copy_fbs():
    source_path = "win/rl/nle/fbs"
    target_path = "nle/fbs"
    if os.path.exists(target_path):
        print("Found existing {} -- Removing old package".format(target_path))
        shutil.rmtree(target_path)
    print("Copying {} to {}".format(source_path, target_path))
    shutil.copytree(source_path, target_path)


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


ext_modules = [
    setuptools.Extension(
        "nle.nethack.helper",
        ["win/rl/helper.cc"],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            "include",
        ],
        language="c++",
        # Warning: This should stay in sync with the Makefiles, otherwise
        # off-by-one errors might happen. E.g. there's a MAIL deamon only
        # if the MAIL macro is set.
        extra_compile_args=["-DNOCLIPPING", "-DNOMAIL", "-DNOTPARMDECL"],
        # This requires `make`ing NetHack before.
        extra_link_args=["src/monst.o", "src/decl.o", "src/drawing.o"],
    )
]


class BuildExt(setuptools.command.build_ext.build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {"msvc": ["/EHsc"], "unix": []}

    if sys.platform == "darwin":
        c_opts["unix"] += ["-stdlib=libc++", "-mmacosx-version-min=10.14"]

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append("-std=c++14")
            opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args += opts
            if sys.platform == "darwin":
                ext.extra_link_args += ["-stdlib=libc++"]

        super().build_extensions()


class Build(distutils.command.build.build):
    """Customized distutils build command."""

    def run(self):
        self.execute(make_install_nethack, [], msg="Building and installing NetHack...")
        super().run()


class Install(setuptools.command.install.install):
    """Customized setuptools develop command."""

    def run(self):
        self.execute(make_install_nethack, [], msg="Building and installing NetHack...")
        super().run()


class Develop(setuptools.command.develop.develop):
    """Customized setuptools develop command."""

    def run(self):
        self.execute(make_install_nethack, [], msg="Building and installing NetHack...")
        super().run()


packages = [
    "nle",
    "nle.env",
    "nle.nethack",
    "nle.fbs",
    "nle.agent",
    "nle.scripts",
    "nle.tests",
]

entry_points = {
    "console_scripts": [
        "nle-play = nle.scripts.play:main",
        "nle-ttyrec = nle.scripts.ttyrec:main",
        "nle-ttyplay = nle.scripts.ttyplay:main",
    ]
}


extras_deps = {
    "dev": [
        "pre-commit>=2.0.1",
        "black>=19.10b0",
        "flake8>=3.7",
        "flake8-bugbear>=20.1",
        "pytest>=5.3",
        "sphinx>=2.4.4",
        "sphinx-rtd-theme==0.4.3",
    ],
    "agent": ["torch>=1.3.1"],
}

extras_deps["all"] = [item for group in extras_deps.values() for item in group]


if __name__ == "__main__":
    package_name = os.getenv("NLE_PACKAGE_NAME", "nle")
    cwd = os.path.dirname(os.path.abspath(__file__))
    version = open("version.txt", "r").read().strip()
    sha = "Unknown"

    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        pass

    if sha != "Unknown" and not os.getenv("NLE_RELEASE_BUILD"):
        version += "+" + sha[:7]
    print("Building wheel {}-{}".format(package_name, version))

    version_path = os.path.join(cwd, "nle", "version.py")
    with open(version_path, "w") as f:
        f.write("__version__ = '{}'\n".format(version))
        f.write("git_version = {}\n".format(repr(sha)))

    # nle.fbs is declared as package, and it is hard to make sure that we
    # override all behaviour of pip [build_py|install|develop|...], without
    # hitting corner cases. So we _always_ build the python fbs files and copy
    # them into nle/fbs. This does mean that effectively setup.py now depends
    # on`flatc`, however this is not a terrible constraint, as 95% of operations
    # already needed it.
    build_fbs()
    copy_fbs()

    with open("README.md") as f:
        long_description = f.read()

    setuptools.setup(
        name=package_name,
        version=version,
        description=(
            "The NetHack Learning Environment (NLE): "
            "a reinforcement learning environment based on NetHack"
        ),
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="The NLE Dev Team",
        url="https://github.com/facebookresearch/nle",
        license="NetHack General Public License",
        entry_points=entry_points,
        packages=packages,
        ext_modules=ext_modules,
        cmdclass={
            "build": Build,
            "build_ext": BuildExt,
            "install": Install,
            "develop": Develop,
        },
        setup_requires=["pybind11>=2.2"],
        install_requires=[
            "pybind11>=2.2",
            "numpy>=1.16",
            "gym>=0.15",
            "pyzmq>=19.0.0",
            "flatbuffers>=1.10",
        ],
        extras_require=extras_deps,
        python_requires=">=3.5",
        classifiers=[
            "License :: OSI Approved :: Nethack General Public License",
            "Development Status :: 4 - Beta",
            "Operating System :: POSIX :: Linux",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: C",
            "Programming Language :: C++",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Games/Entertainment",
        ],
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Blstats(object):
    __slots__ = ['_tab']

    # Blstats
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Blstats
    def CursX(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Blstats
    def CursY(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Blstats
    def StrengthPercentage(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Blstats
    def Strength(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # Blstats
    def Dexterity(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # Blstats
    def Constitution(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(20))
    # Blstats
    def Intelligence(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(24))
    # Blstats
    def Wisdom(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(28))
    # Blstats
    def Charisma(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(32))
    # Blstats
    def Score(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(36))
    # Blstats
    def Hitpoints(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(40))
    # Blstats
    def MaxHitpoints(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(44))
    # Blstats
    def Depth(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(48))
    # Blstats
    def Gold(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(52))
    # Blstats
    def Energy(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(56))
    # Blstats
    def MaxEnergy(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(60))
    # Blstats
    def ArmorClass(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(64))
    # Blstats
    def MonsterLevel(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(68))
    # Blstats
    def ExperienceLevel(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(72))
    # Blstats
    def ExperiencePoints(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(76))
    # Blstats
    def Time(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(80))
    # Blstats
    def HungerState(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(84))
    # Blstats
    def CarryingCapacity(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(88))

def CreateBlstats(builder, cursX, cursY, strengthPercentage, strength, dexterity, constitution, intelligence, wisdom, charisma, score, hitpoints, maxHitpoints, depth, gold, energy, maxEnergy, armorClass, monsterLevel, experienceLevel, experiencePoints, time, hungerState, carryingCapacity):
    builder.Prep(4, 92)
    builder.PrependInt32(carryingCapacity)
    builder.PrependInt32(hungerState)
    builder.PrependInt32(time)
    builder.PrependInt32(experiencePoints)
    builder.PrependInt32(experienceLevel)
    builder.PrependInt32(monsterLevel)
    builder.PrependInt32(armorClass)
    builder.PrependInt32(maxEnergy)
    builder.PrependInt32(energy)
    builder.PrependInt32(gold)
    builder.PrependInt32(depth)
    builder.PrependInt32(maxHitpoints)
    builder.PrependInt32(hitpoints)
    builder.PrependInt32(score)
    builder.PrependInt32(charisma)
    builder.PrependInt32(wisdom)
    builder.PrependInt32(intelligence)
    builder.PrependInt32(constitution)
    builder.PrependInt32(dexterity)
    builder.PrependInt32(strength)
    builder.PrependInt32(strengthPercentage)
    builder.PrependInt32(cursY)
    builder.PrependInt32(cursX)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class You(object):
    __slots__ = ['_tab']

    # You
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # You
    def Ux(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # You
    def Uy(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))
    # You
    def Ux0(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # You
    def Uy0(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3))
    # You
    def Uz(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 4)
        return obj

    # You
    def Uz0(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 6)
        return obj

    # You
    def Uhunger(self): return self._tab.Get(flatbuffers.number_types.Int16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateYou(builder, ux, uy, ux0, uy0, uz_dnum, uz_dlevel, uz0_dnum, uz0_dlevel, uhunger):
    builder.Prep(2, 10)
    builder.PrependInt16(uhunger)
    builder.Prep(1, 2)
    builder.PrependInt8(uz0_dlevel)
    builder.PrependInt8(uz0_dnum)
    builder.Prep(1, 2)
    builder.PrependInt8(uz_dlevel)
    builder.PrependInt8(uz_dnum)
    builder.PrependInt8(uy0)
    builder.PrependInt8(ux0)
    builder.PrependInt8(uy)
    builder.PrependInt8(ux)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Seeds(object):
    __slots__ = ['_tab']

    # Seeds
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Seeds
    def Core(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Seeds
    def Disp(self): return self._tab.Get(flatbuffers.number_types.Uint64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateSeeds(builder, core, disp):
    builder.Prep(8, 16)
    builder.PrependUint64(disp)
    builder.PrependUint64(core)
    return builder.Offset()



# Copyright (c) Facebook, Inc. and its affiliates.
import multiprocessing as mp
import unittest

from nle import nethack
import zmq


NUM_SUBPROCESSES = 2


def _run_nethack():
    env = nethack.NetHack(archivefile=None)
    env.reset()
    env.step(0)
    env.step(0)
    env.step(0)


def _run_nethack_in_subprocesses(num_procs=1):
    ctx = mp.get_context("fork")

    processes = []

    for _ in range(num_procs):
        process = ctx.Process(target=_run_nethack)
        process.start()
        processes.append(process)

    for process in processes:
        process.join()

    return [process.exitcode for process in processes]


class ForkingNetHackTest(unittest.TestCase):
    def test_forking_without_nethack_in_parent(self, num_procs=NUM_SUBPROCESSES):
        self.assertEqual(_run_nethack_in_subprocesses(num_procs), [0] * num_procs)

    def test_forking_with_nethack_in_parent_new_context(
        self, num_procs=NUM_SUBPROCESSES
    ):
        env = nethack.NetHack(archivefile=None, context=zmq.Context())  # noqa: F841
        self.assertEqual(_run_nethack_in_subprocesses(num_procs), [0] * num_procs)

    def test_forking_with_nethack_in_parent(self, num_procs=NUM_SUBPROCESSES):
        # Breaks for pyzmq <= 18.0.0, fixed in
        # https://github.com/zeromq/pyzmq/commit/28c2a36836fc45c09ede4d9962498db449b642d1 # noqa
        env = nethack.NetHack(archivefile=None)  # noqa: F841
        self.assertEqual(_run_nethack_in_subprocesses(num_procs), [0] * num_procs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates.
import fcntl
import os
import signal
import struct
import termios
import threading
import time
import weakref


def _kill(fd, pid, thread):
    os.kill(pid, signal.SIGTERM)
    os.close(fd)
    thread.join()
    os.waitpid(pid, 0)


def _write_frame(fd, buf, channel=0):
    """Write one frame in ttyrec2 format to fd."""
    try:
        usec = time.time_ns() // 1000
    except AttributeError:  # time_ns is only available in Python 3.7+.
        usec = int(1000 * 1000 * time.time())
    sec = usec // 1000 // 1000
    usec -= 1000 * 1000 * sec  # now = sec + usec * 1e-6
    os.write(fd, struct.pack("<iiiB", sec, usec, len(buf), channel))
    os.write(fd, buf)


def _thread_target(fd, record, lock):
    try:
        while True:
            buf = os.read(fd, 1024)
            if not buf:
                break
            with lock:
                _write_frame(record, buf)
    except IOError:
        pass
    os.close(record)


class PtyProcess:
    def __init__(
        self,
        target,
        recordname="process.%(time)s.%(pid)i.ttyrec",
        recordclosefn=lambda rn: None,
    ):
        self._target = target
        self.pid = None
        self.fd = None
        self._recordname = recordname
        self._recordfd = None
        self._recordclosefn = recordclosefn
        self._finalizer = None
        self._lock = threading.Lock()

    def write_record_frame(self, buf, channel):
        with self._lock:
            _write_frame(self._recordfd, buf, channel)

    def write(self, buf):
        self.write_record_frame(buf, 1)
        os.write(self.fd, buf)

    def fork(self, rows=40, columns=80, wait_for_output=False):
        self.pid, self.fd = os.forkpty()
        if self.pid == 0:
            # Both stdout and stderr are attached to fd. We won't be seeing errors.
            # Could be fixed with
            #   fdmaster, fdslave = os.openpty()
            #   # Fork & in child:
            #   os.close(fdmaster)
            #   os.setsid()
            #   fcntl.ioctl(fdslave, termios.TIOCSCTTY)
            #   os.dup2(fdslave, 0)
            #   os.dup2(fdslave, 1)
            self._target()
            os._exit(1)
        fcntl.ioctl(
            self.fd, termios.TIOCSWINSZ, struct.pack("HHHH", rows, columns, 0, 0)
        )
        if self._recordname is None:
            self.filename = None
            self._recordfd = os.open(os.devnull, os.O_WRONLY)
        else:
            self.filename = self._recordname % {
                "time": time.strftime("%Y%m%d-%H%M%S"),
                "pid": self.pid,
            }
            self._recordfd = os.open(
                self.filename, os.O_WRONLY | os.O_CREAT, mode=0o644
            )

        if wait_for_output:
            # Wait for child process to output something before returning.
            buf = os.read(self.fd, 1024)
            if not buf:
                raise RuntimeError("Could not read from child process")
            _write_frame(self._recordfd, buf)

        # Need to be careful not to keep any reference to self.
        thread = threading.Thread(
            target=_thread_target,
            args=(self.fd, self._recordfd, self._lock),
            daemon=True,  # Daemon so we can join in finalizer when exception thrown.
        )
        weakref.finalize(thread, self._recordclosefn, self.filename)
        thread.start()
        self._finalizer = weakref.finalize(self, _kill, self.fd, self.pid, thread)

    def term(self):
        if self._finalizer is not None:
            self._finalizer()
